﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Anwendung                                                                           //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2021 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Cy4Net.h"
#include "Cy4TypePlate.h"
#include "Storing.h"
#include "Board.h"
#include "Timer.h"
#include "Application.h"
#include "String.h"
#include "Main.h"
#include "Mode.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Dienst Anfragebearbeiter  ////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: InfoLedSignal (read write)                                                  //
//                                                                                      //
//  Beschreibung: Legt die Aufgabe der grünen Info-LEDs fest. Folgende Lichtsignale     //
//                können eingestellt werden:                                            //
//                                                                                      //
//                       0 = Kein Signal, Ausgeschaltet.                                //
//                       1 = Leuchtet durchgehend.                                      //
//                       2 = Herzschlag-Signal.                                         //
//                       3 = Zeigt Aktivität auf dem Bus.                               //
//                       x = Wirkungslos.                                               //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |SIG|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [SIG] Lichtsignale der Info-LED, Voreingestellt ist 2.                              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnInfoLedSignal(const Cy4Service* service, Cy4Message* message)
{
	UInt08 data[1];
	{   InfoLedSignal signal;
		STORING_Load(STO_InfoLedSignal, &signal);
		data[0] = (UInt08)signal;
	}
	CY4SERVICE_TransferPayload(service, message, data);
	if (data[0] < 4)
	{   InfoLedSignal signal = (InfoLedSignal)data[0];
		STORING_Save(&signal, STO_InfoLedSignal);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DeviceLabel (read write)                                                    //
//                                                                                      //
//  Beschreibung: Benutzerdefinierte Gerätebezeichnung.                                 //
//                                                                                      //
//              0   1   2   ...  59  60                                                 //
//            +---+---+---+-   -+---+---+                                               //
//  Belegung: |N01|N02|N03| ... |N60| 0 |                                               //
//            +---+---+---+-   -+---+---+                                               //
//                                                                                      //
//  [Nxx] Zeichen der allgemeinen Gerätebezeichnung: 0 .. 255 (ASCII/UTF-8)             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDeviceLabel(const Cy4Service* service, Cy4Message* message)
{
    UInt08  label[61];

    STORING_Load(STO_DeviceLabel, label);
    CY4SERVICE_TransferPayload(service, message, label);
    label[sizeof(label)-1] = '\0';
    STORING_Save(label, STO_DeviceLabel);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: FactorySettings (write only)                                                //
//                                                                                      //
//  Beschreibung: Zurückstellen aller Einstellungen auf den Ursprungszustand.           //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |FST|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [FST] Auslöser zur Wiederherstellung der Werksvoreinstellungen: 5Eh ('^').          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnFactorySettings(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) STORING_RestoreDefaults();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: CallTypePlate (read only)                                            //
//                                                                                      //
//  Beschreibung: Geräte-Typenschild erfragen.                                          //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |AT2|AT1|AT0|AVE|BT2|BT1|BT0|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//  [ATx] Anwendungstypennummer: Axxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//  [AVE] Anwendungsversion:     01 .. 99, Haupt/Unternummer in BCD (Bsp: 12h = V1.2)   //
//  [BTx] Boardtypennummer:      Bxxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//                                                                                      //
//                                                                                      //
//      Beispiel:                                                                       //
//                                                                                      //
//          +---+---+---+---+---+---+---+                                               //
//          |A0h|01h|23h|45h|B0h|67h|89h|      Anwendungstypennummer: 123               //
//          +---+---+---+---+---+---+---+      Anwendungsversion:     4.5               //
//                                             Boardtypennummer:      6789              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnCallTypePlate(const Cy4Service* service, Cy4Message* message)
{
    static const UInt08  typePlate[][7] PROGMEM = {  { 'C','y','4','C','o','d','e' },
                                                     CY4TYPEPLATE(DEV_APPL_TYPE,
                                                                  DEV_APPL_VERS,
                                                                  DEV_BOARD_TYPE)
                                                  };

    UInt08 j = (UInt08)(CY4MESSAGE_Location(message) - CY4SERVICE_Base(service));
    for (UInt08 i=0; i<CY4MESSAGE_Count(message); i++)
    {   CY4MESSAGE_Data(message)[i] = pgm_read_byte(&typePlate[1][j]);
        j++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterApplication (write only)                                        //
//                                                                                      //
//  Beschreibung: Startet die Anwendung neu.                                            //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EAP|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EAP] Auslöser zum (Neu-)Start der Applikation (Reset): 5Eh ('^').                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterApplication(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_Reset();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterProgramLoader (write only)                                      //
//                                                                                      //
//  Beschreibung: Startet das System neu und kehrt im Programmlader-Modul zurück.       //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EPL|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EPL] Auslöser zum Eintritt in den Programmlader: 5Eh ('^').                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterProgramLoader(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_EnterPgmLdr();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: ChangeCy4Address (write only)                                        //
//                                                                                      //
//  Beschreibung: Wechsel der Geräteadresse.                                            //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |NET|DEV|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [NET] Neue Netzadresse:   1 .. 255                                                  //
//  [DEV] Neue Geräteadresse: 1 .. 255                                                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnChangeCy4Address(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[2];
    {   Cy4Address address = CY4NET_Address();
        data[0] = CY4ADDRESS_Net(&address);
        data[1] = CY4ADDRESS_Dev(&address);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    {   Cy4Address address = CY4ADDRESS(data[0], data[1]);
        if (CY4ADDRESS_IsBroadcast(&address) || !CY4ADDRESS_IsValid(&address))
        {   CY4MESSAGE_Invalidate(message);
        }
        else
        {   CY4NET_SetAddress(address);
            STORING_SaveCy4SelfAdr(address);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedLedSignaling(void)
{
	// Info-LED
	InfoLedSignal signal;
	STORING_Load(STO_InfoLedSignal, &signal);
	UInt16 ticks = BOARD_MSecTicks();
	bool ledOn;
	switch (signal)
	{   case INFOLEDSIGNAL_OFF:          ledOn = false;                  break;
		case INFOLEDSIGNAL_ON:           ledOn = true;                   break;
		case INFOLEDSIGNAL_HEARTBEAT:    ledOn = (bool)((ticks & 0x380)==0x80 || (ticks & 0x380)==0x180); break;
		case INFOLEDSIGNAL_BUS_ACTIVITY: ledOn = BOARD_BusActivity();    break;
		default:                         ledOn = false;
	}
	BOARD_SetLed(ledOn);	
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Startup(void)
{
    static const Cy4ServiceItem  servicesTable[] PROGMEM =
    {   //|                     service                     |        handler        |
        //+------------------+-----+-----------------.....--+-----------------------+        
		{ CY4SERVICE(0xFFF020,  1  , CY4SERVICE_READ_WRITE), OnInfoLedSignal        },
        { CY4SERVICE(0xFFFF00, 61  , CY4SERVICE_READ_WRITE), OnDeviceLabel          },
        { CY4SERVICE(0xFFFF40,  1  , CY4SERVICE_WRITE_ONLY), OnFactorySettings      },
        { CY4SERVICE(0xFFFFF0,  7  , CY4SERVICE_READ_ONLY ), OnCallTypePlate        },
        { CY4SERVICE(0xFFFFF8,  1  , CY4SERVICE_WRITE_ONLY), OnEnterApplication     },
        { CY4SERVICE(0xFFFFFA,  1  , CY4SERVICE_WRITE_ONLY), OnEnterProgramLoader   },
        { CY4SERVICE(0xFFFFFC,  2  , CY4SERVICE_WRITE_ONLY), OnChangeCy4Address     }
    };
    CY4NET_AssignServices(servicesTable);
    Cy4Address address = STORING_LoadCy4SelfAdr();
    CY4NET_SetAddress(address);       
    MODE_ShowStartInfo();    
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Proceed(void)
{
    ProceedLedSignaling();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Execute(void)
{
    while (MAIN_DoIdleProcessing())
    {   bool forceRedraw = false;       // 'true' Anzeige der Leitungsspannung forcieren
        // Tastendrücke verarbeiten
        static Timer pressDuration = TIMER_MILLISECONDS(CFG_ENTER_SETTINGS_DELAY);
        static bool  pressed       = false;
        if (pressed)
        {   if (BOARD_IsPushButtonPressed(PUSHBUTTON_ENCODER))
            {   if (TIMER_HasExpired(&pressDuration))
                {   // Drehinkrement-Taste wurde längere Zeit gedrückt gehalten
                    MODE_SetOwnAddress();
                    pressed = false;
                    forceRedraw = true;
                }
            }
            else
            {	if (TIMER_Elapsed(&pressDuration) < 1000)
				{	// Drehinkrement-Taste wurde nur kurz gedrückt gehalten
					MODE_SetDestinationAddress();
					forceRedraw = true;
				}
                pressed = false;                
            }
        }
        else if (BOARD_TakePushButton(PUSHBUTTON_ENCODER))
        {   pressed = true;
            TIMER_Restart(&pressDuration);
        }
        else if (BOARD_TakePushButton(PUSHBUTTON_DIRECT))
        {   // Freie Taste wurde gedrückt
            MODE_SendPingSequence();
            forceRedraw = true;
        }
        // Aktuelle Leitungsspannung holen
        SInt16 voltage = BOARD_GetLineVoltage();
        static SInt16 prevVoltage = INVALID;
        if( prevVoltage != voltage )
        {   prevVoltage  = voltage;
            forceRedraw = true;
        }
        // Leitungsspannung anzeigen
        if (forceRedraw)
        {   String string = STRING_EMPTY;
            STRING_PrintText(&string,"\rU: ");
            if( voltage < 0 || voltage > 999 )
            {   STRING_PrintText(&string," x.x");
            }
            else
            {   if( voltage < 100 ) STRING_PrintChar(&string,' ');
                STRING_PrintFixpoint(&string, voltage, 1);
            }
            STRING_PrintChar(&string, 'V');
            BOARD_PrintOnDisplay(STRING_ConstData(&string));
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
