﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gerätespezifischer Funktionen und Mechanismen            //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2021 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "AVRChip.h"
#include "Board.h"
#include "Timer.h"
#include "RingBuffer.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define DSPL_CSB    B,0
#define DSPL_RS     B,1
#define DSPL_RESET  B,2
#define DIRECT_PB   D,2
#define ROT_ENC_PB  D,3
#define ROT_ENC_A   D,4
#define ROT_ENC_B   D,5
#define SYSTEM_LED  D,6

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                        // Arbeitsanweisungen für das Board-Modul:
{   DO_STARTUP,                     // · Starte die Anwendung neu
    DO_ENTER_PGMLDR,                // · Starte den Programmlader
    DO_PROCEED,                     // · Führe die Arbeit fort
    DO_SHUTDOWN                     // · Schalte das Board ab
} DoNext;

typedef struct                      // Bus-Übertragungsstatus:
{   bool   pending : 1;             // · Übertragung im Gange
    bool   echo    : 1;             // · Echo erhalten
    bool   error   : 1;             // · Übertragung Fehlgeschlagen (Kollision)
    UInt08 byteGap : 5;             // · Vergangene Zeit seit dem letzten Byteempfang [ms]
} BusStatus;

typedef struct                      // Element der Bus-Empfangsschlange:
{   bool   timeout : 1;             // · Zeitüberschreitung bei <byteGap> aufgetreten
    UInt08 padding : 7;             // · Füllbits (ohne Bedeutung)
    UInt08 data    : 8;             // · Datenbyte
} BusRxItem;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static volatile UInt16     msecTicks                             = 0;
static volatile UInt16     secTicks                              = 0;
static volatile UInt08     busTxQueue[CFG_BUS_TX_QUEUE_CAPACITY] = { 0 };
static volatile UInt08     busTxQueueHead                        = 0;
static volatile UInt08     busTxQueueLength                      = 0;
static volatile UInt08     busLoopbackByte                       = 0;
static volatile BusStatus  busStatus                             = { false, false, false, 0};
static volatile BusRxItem  busRxQueue[CFG_BUS_RX_QUEUE_CAPACITY] = { { false } };
static volatile UInt08     busRxQueueHead                        = 0;
static volatile UInt08     busRxQueueTail                        = 0;
static          UInt08     busActivityIndicator                  = 0;
static          bool       pushButtonPressedEvent[2]             = { false, false };
static          bool       pushButtonIsPressed[2]                = { false, false };
static volatile SInt08     encoderPosition                       = 0;
static          SInt16     lineVoltage                           = INVALID;
static          DoNext     doNext                                = DO_STARTUP;

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_PreStart(): Diese Funktion wird einmal nach dem Systemstart aufgerufen. Der   //
//  ----------------  Zeitpunkt ihres Aufrufs macht sie zur einer Sonderfunktion, denn  //
//  ihr Aufruf findet in der Start-Sektion "init3", noch lange vor Aufruf von main(),   //
//  statt. Das ist der richtige Zeitpunkt, um z.B. die Ursache des Resets zu ermitteln  //
//  oder statische Variablen auszulesen, deren Werte vor dem Reset platziert wurden     //
//  und den Neustart überlebt haben. Denn zu diesem Zeitpunkt sind die statischen       //
//  Variablen noch nicht initialisiert.                                                 //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked, section(".init3"))) void
BOARD_PreStart(void)
{
    MCUSR = 0;
    wdt_disable();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  Reset(): Führt einen physikalischen Reset herbei indem der Watchdog ausgelöst wird. //
//  -------  Reset() kehrt nie zum Aufrufer zurück.                                     //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) static void
Reset(void)
{
    DisableInterrupts();            // Interrupts sperren
    wdt_enable(WDTO_15MS);          // Watchdog-Intervall auf kürzeste Zeit setzen
    __asm__ volatile
    (   "ldi r30, lo8 (%0) \n"      // Nach einem Neustart des Systems wird als
        "ldi r31, hi8 (%0) \n"      // erstes das Programmladermodul ausgeführt.
        "ldi r24, %1       \n"      // Der Programmlader nutzt zur Steuerung seines
        "ldi r25, hi8 (%2) \n"      // Ablaufs Variablen, die in der "Noinit"-Sektion
        "0:                \n"      // angesiedelt sind, d.h. sie werden im Rahmen
        "st  Z+,  r24      \n"      // eines Neustartes nicht initialisiert.
        "cpi r30, lo8 (%2) \n"      // Um nun zu vermeiden, dass übrig gebliebene
        "cpc r31, r25      \n"      // Werte des SRAMs unerwünschte Effekte im
        "brlo 0b"                   // Programmlader hervorrufen, wird vor dem
            :                       // Reset das komplette SRAM mit Nullen überschrieben.
            : "i" (RAMSTART),       // Da durch die Nullung Heap wie Stack komplett
              "i" (0),              // zerstört sind, ist diese Aktion ein Punkt-Ohne-
              "i" (RAMEND+1)        // Wiederkehr.
    );
    while (true);                   // Warten, bis Watchdog-Reset eintritt
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
EnterPgmLdr(void)
{
    if (pgm_read_word(FLASHEND - CFG_PGMLDR_SIZE*1024 + 1) == 0x940C)
    {   GOTO(FLASHEND - CFG_PGMLDR_SIZE*1024 + 1 + 4);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
FreezeMs(UInt08 ms)
{
    for (UInt16 c=0; c<ms*10; c++) Freeze100us();
}

//////////////////////////////////////////////////////////////////////////////////////////
static SInt16
ReadADC(UInt08 channel)
{    
    ADMUX = BV(REFS0) | channel;  // Eingangsspannung bezogen AVCC (5V), Kanalwahl: 0 .. 7
    ADCSRA |= BV(ADSC);
    loop_until_bit_is_clear(ADCSRA, ADSC);
    UInt08 lo = ADCL;
    UInt08 hi = ADCH;
    SInt16 adcVal = (SInt16)ToWord(hi, lo);   // 0 .. 5 [V] -> 0 .. 1023 [#]
    if (adcVal < 0 || adcVal > 1023) adcVal = INVALID;
    return adcVal;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
TransmitSPI(UInt08 data)
{
    SPDR = data;
    loop_until_bit_is_set(SPSR, SPIF);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
LCDSendCmd(UInt08 cmd)
{
    ClrPin(DSPL_RS);    // RS = L (Kommando)
    ClrPin(DSPL_CSB);   // ¬CSB = L (Chip select)
    TransmitSPI(cmd);   // Byte via SPI übertragen
    SetPin(DSPL_CSB);   // ¬CSB = H (Chip deselect)
    Freeze100us();
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
LCDSendData(UInt08 data)
{
    SetPin(DSPL_RS);    // RS = H (Daten)
    ClrPin(DSPL_CSB);   // ¬CSB = L (Chip select)
    TransmitSPI(data);  // Byte via SPI übertragen
    SetPin(DSPL_CSB);   // ¬CSB = H (Chip deselect)
    Freeze100us();
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
LCDPrintChar(char c)
{
    static UInt08  cursorPos = 0;

    switch (c)
    {   case '\r':      // Befehl "Cursor home"
			LCDSendCmd(0x02);
			cursorPos = 0;
			FreezeMs(5);
			break;
        case '\n':      // Befehl "Clear Display"
			LCDSendCmd(0x01);
			cursorPos = 0;
			FreezeMs(5);
			break;
        default:
			if (c >= '\x10' && c <= '\x17') c -= '\x10';
			if (cursorPos == 0) LCDSendCmd(0x80);
			if (cursorPos < 8)
			{   LCDSendData(c);
				cursorPos++;
			}
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedLineVoltageMeasuring(void)
{
    static Timer  update = TIMER_MILLISECONDS(CFG_VOLTAGE_READOUT_INTERVAL+1);
    if (!TIMER_RestartIfExpired(&update)) return;    
    // (1) ADC Auslesen
	SInt16 adc = ReadADC(1);     // Leitungsspannung auf Kanal 1: 0 .. 1023
	// (2) Spannungswert ermitteln
    if (adc == INVALID)
	{	lineVoltage = INVALID;  
	}
	else	      
	{	lineVoltage = (SInt16)(((SInt32)adc*3300)/10230)+8; // +0.8V (D2)
		lineVoltage += CFG_VOLTAGE_ADJUSTMENT;		
		if (lineVoltage < 0) lineVoltage = 0;		
	}
	// (3) Spannungswert durch Medianfilter glätten
	static SInt16     ringBufferData[CFG_VOLTAGE_RINGBUFFER_CAPACITY];
	static RingBuffer ringBuffer = RINGBUFFER_INIT(ringBufferData);
	RINGBUFFER_Add(&ringBuffer, lineVoltage);
	lineVoltage = RINGBUFFER_Median(&ringBuffer); // 0 .. 330  [V÷10]
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedPushButtons(void)
{
    // 1ms Zeitbarriere
    UInt16 ticks = BOARD_MSecTicks();
    {   static UInt08 ms = 0;
        if (ms == (UInt08)ticks) return;
        ms = (UInt08)ticks;
    }
    // Tastenkontrolle Alleinstehende Drucktaste (0) und Inkrementgeber-Drucktaste (1)
    for (UInt08 i=0; i<2; i++)
    {   static UInt08 latency[2] = { 0, 0 };
        bool isPressed;
        switch (i)
        {   case 0:  isPressed = !GetPin(DIRECT_PB);  break;
            case 1:  isPressed = !GetPin(ROT_ENC_PB); break;
            default: isPressed = false;
        }
        if (pushButtonIsPressed[i])
        {   // Taster geschlossen
            if (!isPressed && latency[i] > 0) latency[i]--;
            else                              latency[i] = CFG_PUSHBUTTON_UP_RESPONSIVITY;
        }
        else
        {   // Taster geöffnet
            if (isPressed && latency[i] > 0) latency[i]--;
            else                             latency[i] = CFG_PUSHBUTTON_DOWN_RESPONSIVITY;
        }
        if (latency[i] == 0)
        {   pushButtonIsPressed[i]    = isPressed;
            pushButtonPressedEvent[i] = isPressed;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedIndicators(void)
{
	// 1ms Zeitbarriere
	UInt16 ticks = BOARD_MSecTicks();
	{	static UInt08 ms = 0;
		if (ms == (UInt08)ticks) return;
		ms = (UInt08)ticks;
	}
	// Bus-Aktivitätsindikator
	{   if (busActivityIndicator > 0)
		{   busActivityIndicator--;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(TIMER2_COMPA_vect)
{
    // Ticker inkrementieren
    {   msecTicks++;
        static UInt16 counter = 0;
        counter++;
        if (counter == 1000)
        {   counter = 0;
            secTicks++;
        }
    }
    // Drehinkrementalgeber (s. https://www.grasbon.de/artikel/Drehinkrementalgeber/index.html)
    {
        #if CFG_ROTARY_ENCODER_CHANNELS == 2
        // ... für Inkrementalgeber mit Auswertung der Signale A und B (z.B. DDM Hopt+Schuler)
        static const SInt08 encZt[][4] PROGMEM = { {  0, -1, +1,  0 },
                                                   { +1,  0,  0, -1 },
                                                   { -1,  0,  0, +1 },
                                                   {  0, +1, -1,  0 }
                                                 };
        #elif CFG_ROTARY_ENCODER_CHANNELS == 1
        // ... für Inkrementalgeber mit Auswertung nur des A-Signals (z.B. Alps)
        static const SInt08 encZt[][4] PROGMEM = { {  0,  0, +1,  0 },
                                                   {  0,  0,  0, -1 },
                                                   { -1,  0,  0,  0 },
                                                   {  0, +1,  0,  0 }
                                                 };
        #else
            #error Undefined CFG_ROTARY_ENCODER_CHANNELS. Must be 1 or 2
        #endif
        
        static UInt08  state = 0;
        static SInt08  delta = 0;
        
        UInt08 ab = (!GetPin(ROT_ENC_A) << 1) + !GetPin(ROT_ENC_B);
        delta += (SInt08)pgm_read_byte(&(encZt[state][ab]));
        state = ab;
        if (delta ==  CFG_ROTARY_ENCODER_RASTER_STOP ||
            delta == -CFG_ROTARY_ENCODER_RASTER_STOP)
        {   delta > 0 ? encoderPosition++ : encoderPosition--;
            delta = 0;
        }        
    }
    // Bus Byte-Zeitlücke
    if (busStatus.byteGap <= CFG_BUS_INTERBYTE_TIMEOUT)
    {   busStatus.byteGap++;
        if (busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT)
        {   busRxQueue[busRxQueueTail].timeout = true;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART_TX_vect)
{                                   // IRQ Abfolge sowie die Signale <echo> und <pending>:
	if (!busStatus.echo)            //
	{   busStatus.error = true;     // Start Tx:    |1|
		busTxQueueLength = 0;       //               :
	}                               // Tx-IRQ:  ···--:-----|2|----|3|-···-|n|----| |-···
	busStatus.echo = false;         // Rx-IRQ:  ···--:-|1|--:-|2|--:--···-- -|n|--:--···
	UInt08 head = busTxQueueHead;   //               :  :___:  :___:          :___:
	if (busTxQueueLength == 0)      // echo:    ...__:__|   |__|   |____...___|   |__...
	{   busStatus.pending = false;  //               :____________________________:
	}                               // pending: ...__|                            |__...
	else
	{   busLoopbackByte = busTxQueue[head];
		UDR0            = busTxQueue[head];
		head++;
		if (head == CARDINALITY(busTxQueue)) head = 0;
		busTxQueueLength--;
	}
	busTxQueueHead = head;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART_RX_vect)
{
	UInt08 byte = UDR0;
	busStatus.byteGap = 0;   // Byte empfangen, Zeitzähler zurücksetzen
	if (busStatus.pending && byte == busLoopbackByte) busStatus.echo = true; // Echo
	// Byte in Empfangsschlange stellen
	UInt08 tail = busRxQueueTail;
	busRxQueue[tail].data = byte;
	tail++;
	if (tail == CARDINALITY(busRxQueue)) tail = 0;
	if (tail == busRxQueueHead) busRxQueue[busRxQueueTail].timeout = true; // Überlauf
	else                        busRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Reset(): Initiiert einen Systemneustart. Die Funktion löst nicht unmittelbar  //
//  -------------  einen Reset aus, sondern führt zu einem Ausstieg aus der Haupt-      //
//  schleife main(). Die danach sukzessiv aufgerufenen "Shutdown"-Funktionen aller      //
//  betroffenen Module bewirken den kontrollierten Abschluss aller Aktivitäten, bevor   //
//  die letzte aufgerufene "Shutdown"-Funktion des Board-Moduls dann den physikali-     //
//  schen Reset auslöst.                                                                //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Reset(void)
{
	doNext = DO_STARTUP;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EnterPgmLdr(): Ähnlich wie BOARD_Reset() initiiert BOARD_EnterPgmLdr() einen  //
//  -------------------  Systemneustart. Im Unterschied zu BOARD_Reset(), wird hier     //
//  jedoch nicht die Anwendung neu gestartet, sondern das System bootet in das Pro-     //
//  grammladermodul.                                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnterPgmLdr(void)
{
	doNext = DO_ENTER_PGMLDR;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_MSecTicks(): Die Funktion liefert den Stand eines monotonen Millisekunden-    //
//  -----------------  Zeitzählers. Der Zähler wird interruptgetrieben inkrementiert    //
//  und hat eine Breite von 16-Bit. Kommt es nach 65535 ms zum Überlauf, beginnt der    //
//  Zeitzähler wieder bei 0.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Millisekunden-Zeitzählers.                             //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_MSecTicks(void)
{
	UInt16 ticks = 0;

	Atomic { ticks = msecTicks; }
	return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SecTicks(): Die Funktion liefert den Stand eines monotonen Sekunden-Zeit-     //
//  ----------------  zählers. Der Zähler wird interruptgetrieben inkrementiert und hat //
//  eine Breite von 16-Bit. Kommt es nach 65535 s (knapp 18¼h) zum Überlauf, beginnt    //
//  der Zeitzähler wieder bei 0.                                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Sekunden-Zeitzählers.                                  //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_SecTicks(void)
{
	UInt16 ticks = 0;

	Atomic { ticks = secTicks; }
	return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPLoad(): Kopiert <count> Bytes aus dem EEPROM-Speicher ins RAM.             //
//  ---------------                                                                     //
//                                                                                      //
//  Eingabe: · eepAddress: Adresse im EEPROM-Speicher.                                  //
//           · ramAddress: Adresse im Ram.                                              //
//           · count:      Anzahl zu kopierender Bytes.                                 //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPLoad(const void* eepAddress, void* ramAddress, UInt16 count)
{
	eeprom_read_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPSave(): Kopiert <count> Bytes aus dem RAM in den EEPROM-Speicher. Da       //
//  ---------------  EEPROM-Speicher lt. Datenblatt nur ca. 100000 Lösch/Schreibzyklen  //
//  pro Zelle verträgt, wird vor für jedes Bytes vor dem Speichervorgang geprüft, ob    //
//  sich der zu schreibende Wert vom Inhalt der EEP-Speicherzelle unterscheidet. Erst   //
//  danach wird nötigenfalls ein Lösch- und Schreibvorgang durchgeführt.                //
//  Hinweis: Das Beschreiben einer EEPROM-Speicherzelle nimmt mit ca. 8 ms verhältnis-  //
//           mäßig viel Zeit in Anspruch.                                               //
//                                                                                      //
//  Eingabe: · ramAddress: Adresse im Ram.                                              //
//           · eepAddress: Adresse im EEPROM-Speicher.                                  //
//           · count:      Anzahl zu kopierender Bytes.                                 //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPSave(const void* ramAddress, void* eepAddress, UInt16 count)
{
	eeprom_update_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsIdle(): Prüft, ob der Bus momentan frei ist. Der Bus gilt als frei, wenn //
//  -----------------  eine Kommunikationspause von mehr als 'CFG_BUS_INTERBYTE_TIME-   //
//  OUT' Millisekunden eintritt.                                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Bus frei ist.                                             //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsIdle(void)
{
	bool  idle = false;
	Atomic
	{   idle = !busStatus.pending && busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT;
	}
	return idle;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTransmit(): Sendet das Datenbyte in <byte> auf den Bus.                    //
//  -------------------  Die Sendeeinheit ist durch folgenden Zustandsautomaten         //
//  realisiert, Ausgangszustand ist IDLE:                                               //
//                                                                                      //
//                                                                                      //
//     +----------+ BOARD_BusTransmit()      +--------------+ BOARD_BusTransmit()       //
//     |//////////+------------------------->|//////////////+--------------------+      //
//     |// IDLE //|                COMPLETED |// TRANSMIT //|                    |      //
//     |//////////|<-------------------------+//////////////|<-------------------+      //
//     +----------+                          +------+-------+                           //
//          /|\                                     | COLLISION                         //
//           |                                      |                                   //
//           |                                     \|/                                  //
//           |                               +-------------+                            //
//           | BOARD_BusTakeFailure()        |/////////////|                            //
//           +-------------------------------+// FAILURE //|                            //
//                                           |/////////////|                            //
//                                           +-------------+                            //
//                                                                                      //
//  Mit Aufruf dieser Funktion wechselt der Automat von IDLE nach TRANSMIT und beginnt  //
//  mit mit der Übertragung des angegebenen Datenbytes. Ist bereits eine Übertragung im //
//  Gange (Zustand TRANSMIT), wird bei Aufruf der Funktion das zu übertragende Daten-   //
//  byte für die spätere Übertragung in die Sende-Warteschlange gestellt. Nach Aus-     //
//  sendung aller zu übertragenden Datenbytes (Ereignis COMPLETED) kehrt der Automat    //
//  wieder in den Zustand IDLE zurück.                                                  //
//  Das Halbduplexverfahren des seriellen Busses bedingt, dass jedes ausgesendete Byte  //
//  unmittelbar als Echo wieder in den Empfangspuffer zurückwandert. Diese Rücküber-    //
//  tragung wird genutzt, um Kollisionen aufzudecken. Unterscheiden sich ausgesendetes  //
//  Byte und Echo, löst das das Ereignis COLLISION aus und damit den Wechsel in den     //
//  Zustand FAILURE. Die Übertragung wird in dem Fall nicht mehr fortgesetzt und die    //
//  Sende-Warteschlange geleert. Gegebenenfalls noch ausstehende Bytes werden also      //
//  nicht mehr übertragen. Im Zustand FAILURE bleiben Aufrufe von BOARD_BusTransmit()   //
//  ohne Wirkung. Um neue Datenbytes senden zu können, muss der Automat erst durch      //
//  Aufruf von BOARD_BusTakeFailure() wieder in den Zustand IDLE versetzt werden.       //
//                                                                                      //
//  Eingabe: · byte: Das zu übertragende Datenbyte.                                     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusTransmit(UInt08 byte)
{
	// Ist die Übertragungsschlange voll, warten, bis wieder Platz ist
	bool  queueFull = false;
	do
	{   Atomic
		{   queueFull = busTxQueueLength == CARDINALITY(busTxQueue);
		}
	} while (queueFull);
	// Byte senden
	Atomic
	{   if (busStatus.error) return; // Zustand FAILURE: Nichts tun
		if (busStatus.pending)
		{   // Zustand TRANSMIT: Byte für Übertragung aufreihen
			UInt08 tail = busTxQueueHead + busTxQueueLength;
			if (tail >= CARDINALITY(busTxQueue)) tail -= CARDINALITY(busTxQueue);
			busTxQueue[tail] = byte;
			busTxQueueLength++;
		}
		else
		{   // Zustand IDLE: Byte unmittelbar senden
			busStatus.pending = true;
			busStatus.echo    = false;
			busLoopbackByte   = byte;
			UDR0              = byte;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsTransmitting(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit())   //
//  -------------------------  im Zustand TRANSMIT befindet. Das ist solange der Fall,  //
//  wie die Sendeeinheit mit der Übertragung von Datenbytes beschäftigt ist.            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Sendevorgang noch im Gange ist.                           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsTransmitting(void)
{
	bool  pending = false;
	Atomic
	{   pending = busStatus.pending;
	}
	return pending;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTakeFailure(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit()) im   //
//  ----------------------  Zustand FAILURE befindet und löst gegebenenfalls einen      //
//  Wechsel in den Zustand IDLE aus. Übertragungsfehler entstehen in der Regel durch    //
//  Kollisionen auf dem Bus.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn ein Übertragungsfehler vorlag.                                //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusTakeFailure(void)
{
	bool  error = false;
	Atomic
	{   error = busStatus.error;
		busStatus.error = false;
	}
	return error;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusReceive(): Steht bei Aufruf der Funktion ein empfangenes Datenbyte zur     //
//  ------------------  Abholung bereit, wird es aus der Empfangsschlange entfernt und  //
//  in <byte> übergeben. BOARD_BusReceive() kehrt so lange mit 'true' zurück, wie       //
//  Datenbytes zur Abholung bereitstehen.                                               //
//  Man beachte: Bei einer Unterbrechung der Empfangssequenz kehrt die Funktion mit     //
//  'false' zurück. Umgekehrt gilt das jedoch nicht: Der Rückgabewert 'false' bedeutet  //
//  nicht zwingend, dass auch die Empfangssequenz bereits unterbrochen ist! Dazu siehe  //
//  BOARD_BusIsReceiving().                                                             //
//                                                                                      //
//  Eingabe: · byte: Referenz, die das nächste zur Abholung bereitstehende Datenbyte    //
//                   aufnimmt.                                                          //
//  Ausgabe: 'true', wenn in <byte> ein Datenbyte abgelegt wurde.                       //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusReceive(UInt08* byte)
{
	bool  hasData = false;
	Atomic
	{   UInt08 head = busRxQueueHead;
		if (!busRxQueue[head].timeout)
		{   // Keine Zeitüberschreitungs-Markierung gesetzt
			if (head != busRxQueueTail)
			{   // Datenbyte vorhanden
				*byte = busRxQueue[head].data;
				head++;
				if (head == CARDINALITY(busRxQueue)) head = 0;
				busRxQueueHead = head;
				hasData = true;
			}
		}
	}
	return hasData;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsReceiving(): Prüft, ob die aktuelle Empfangssequenz noch im Gange ist.   //
//  ----------------------                                                              //
//  Eine Gruppe empfangener Datenbytes gilt als zusammenhängende Empfangssequenz, wenn  //
//  die Zeitspanne zwischen dem Empfang sukzessiver Datenbytes einen bestimmten Zeit-   //
//  rahmen nicht überschreitet. Mit anderen Worten: Liegen die Zeitpunkte zweier auf-   //
//  einanderfolgend empfangener Datenbytes länger als 'CFG_BUS_INTERBYTE_TIMEOUT'       //
//  Millisekunden auseinander (s.u. timeout), gilt die Empfangssequenz als unter-		//
//	brochen.     																		//
//                                                                                      //
//    Empfang: ···-|A|-|B|-|C|---------------×---|D|----|E|---------------×--|F|-···    //
//                           :               :            :               :             //
//                           |--- timeout ---|            |--- timeout ---|             //
//                                                                                      //
//    Empfangssequenzen: ABC, DE, F...                                                  //
//                                                                                      //
//  Die Unterbrechungen werden mit in der Empfangsschlange vermerkt, so dass beim Aus-  //
//  lesen der empfangenen Bytes mittels BOARD_BusReceive() die Empfangssequenzen zeit-  //
//  unabhängig separiert werden. Solange BOARD_BusIsReceiving() mit 'true' zurückkehrt, //
//  ist die Empfangssequenz noch im Gange.                                              //
//                                                                                      //
//      Beispiel: Empfang von Byte-Sequenzen:                                           //
//                                                                                      //
//       UInt08 sequence[SEQUENCE_CAPACITY];   // SEQUENCE_CAPACITY: 1 .. 255           //
//                                                                                      //
//       UInt08                                                                         //
//       TakeNextSequence(void)                                                         //
//       {                                                                              //
//           static UInt08 index = 0;                                                   //
//                                                                                      //
//           UInt08 length = 0;                                                         //
//           if (BOARD_BusIsReceiving())                                                //
//           {   // Empfang einer Sequenz im Gange                                      //
//               UInt08 byte;                                                           //
//               while (BOARD_BusReceive(&byte)) // Datenbytes abholen und aufsammeln   //
//               {   if (index < sizeof(sequence))                                      //
//                   {   sequence[index] = byte;                                        //
//                       index++;                                                       //
//                   }                                                                  //
//                   else                                                               //
//                   {   // Überlauf ...                                                //
//                   }                                                                  //
//               }                                                                      //
//           }                                                                          //
//           else                                                                       //
//           {    // Empfang einer Sequenz nicht (mehr) im Gange                        //
//                length = index;                                                       //
//                index  = 0;                                                           //
//           }                                                                          //
//           return length;   // Rückgabewert: Aktuelle Sequenzlänge                    //
//       }                                                                              //
//                                                                                      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', solange die Empfangssequenz im Gange ist.                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsReceiving(void)
{
	bool  timeout = false;
	Atomic
	{   UInt08 head = busRxQueueHead;
		timeout = busRxQueue[head].timeout;
		if (head != busRxQueueTail) busRxQueue[head].timeout = false;
	}
	return !timeout;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusOnActivity(): Rückruffunktion der Bus-Übertragungseinheit. Die Funktion    //
//  ---------------------  wird aufgerufen, wenn mit dem Gerät über den Bus             //
//  kommuniziert wird.                                                                  //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusOnActivity(void)
{
	busActivityIndicator = CFG_BUS_ACT_INDICATOR_DURATION;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusActivity(): Der Bus-Aktivitätsindikator zeigt an, ob aktuell mit dem       //
//  -------------------  Gerät kommuniziert wird. Diese Funktion liefert den Zustand    //
//  des Indikators. Nach Abschluss der Kommunikation bleibt der Indikator noch 'CFG_-   //
//  BUS_ACT_INDICATOR_DURATION' Millisekunden aktiviert.                                //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusActivity(void)
{
	return busActivityIndicator != 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_TakePushButton(): Liefert 'true', wenn seit der letzten Abfrage mit BOARD_    //
//  ----------------------  TakePushButton() die Taste <pushButton> erneut gedrückt     //
//                          wurde und aktuell noch gehalten wird.                       //
//                                                                                      //
//  Eingabe: · pushButton: Betroffene Taste gem. Aufzählungstyp <PushButton>.           //
//  Ausgabe: 'true', wenn ein Tastendruck-Ereignis stattgefunden hat.                   //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_TakePushButton(PushButton pushButton)
{
	UInt08 i = (UInt08)pushButton;
    bool ret = pushButtonPressedEvent[i];
    pushButtonPressedEvent[i] = false;
    return ret;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_IsPushButtonPressed(): Liefert den aktuellen Zustand der Taste <pushButton>.  //
//  ---------------------------                                                         //
//                                                                                      //
//  Eingabe: · pushButton: Betroffene Taste gem. Aufzählungstyp <PushButton>.           //
//  Ausgabe: 'true', wenn die Taste gerade gedrückt ist.                                //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_IsPushButtonPressed(PushButton pushButton)
{
    return pushButtonIsPressed[(UInt08)pushButton];
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_TakeEncoder(): Liefert die Veränderung des Drehinkrementgebers seit der       //
//  -------------------  letzten Abfrage mit BOARD_TakeEncoder(). Positive Werte be-    //
//  deuten eine Drehbewegung um entsprechend viele Rasterpositionen im Uhrzeigersinn,   //
//  negative Wert entsprechend umgekehrt. Die Position des Drehinkrementgebers sollte   //
//  so häufig abgefragt werden, dass ein Über-/Unterlauf nicht stattfindet.             //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Veränderung der Drehposition: -127 .. +127 [Rasterpositionen].             //
//////////////////////////////////////////////////////////////////////////////////////////
SInt08
BOARD_TakeEncoder(void)
{
    SInt08 value = 0;

    Atomic
    {   value = encoderPosition;
        encoderPosition = 0;
    }
    return value;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetLed(bool on)
{
    ChgPin(SYSTEM_LED, on);
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_Led(void)
{
    return (bool)GetPin(SYSTEM_LED);
}

//////////////////////////////////////////////////////////////////////////////////////////
SInt16
BOARD_GetLineVoltage(void)
{
    return lineVoltage;     // 0 .. 330 [V÷10] oder INVALID
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_PrintOnDisplay(const char* text)
{
    while( *text != '\0' )
    {   LCDPrintChar(*text);
        text++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Startup(void)
{
    if (doNext != DO_STARTUP) Reset();
    DisableInterrupts();
    // Watchdog --------------------------------------------------------------------------
    wdt_enable(WDTO_2S);
    // Ports'n'Pins ----------------------------------------------------------------------
    ConfPort(B, _H_,    // PB.0: DSPL_CSB   : Chipauswahl,  L = Chip select
                _L_,    // PB.1: DSPL_RS    : Registerwahl, H = Daten, L = Kommando
                _L_,    // PB.2: DSPL_RESET : Zurücksetzen, L = Reset
                _L_,    // PB.3: [MOSI]
                _L_,    // PB.4: [MISO]
                _L_,    // PB.5: [SCK]
                _T_,    // PB.6: [XTAL1]
                _T_);   // PB.7: [XTAL2]
    ConfPort(C, _L_,    // PC.0: ...
                _T_,    // PC.1: ADC Kanal 1 (UMon)
                _L_,    // PC.2: ...
                _L_,    // PC.3: ...
                _L_,    // PC.4: ...
                _L_,    // PC.5: ...
                _T_,    // PC.6: [RESET]
                _T_);   //
    ConfPort(D, _T_,    // PD.0: [RXD]
                _T_,    // PD.1: [TXD]
                _P_,    // PD.2: DIRECT_PB  : Drucktaste, L = gedrückt
                _P_,    // PD.3  ROT_ENC_PB : Drehinkrementgeber-Drucktaste, L = gedrückt
                _P_,    // PD.4: ROT_ENC_A  : Drehinkrementgeber Pin A, L = geschlossen
                _P_,    // PD.5: ROT_ENC_B  : Drehinkrementgeber Pin B, L = geschlossen
                _L_,    // PD.6: SYSTEM_LED: Rote System-LED, L = aus
                _L_);   // PD.7: ...
    // Pause, bis sich alles stabilisiert hat --------------------------------------------
    Freeze100us();
    // ADCs ------------------------------------------------------------------------------
    ADCSRA  = BV(ADEN) | BV(ADPS2) | BV(ADPS1) | BV(ADPS0); // ADCs einschalten (FCLK÷128)
    ADCSRA |= BV(ADSC);                                     // 1. Abtastung anstoßen
    // Timer 2: 1ms (Clear Timer on Compare (CTC), F_CPU÷32 = 115,200 kHz) ---------------
    TCCR2A = BV(WGM21);
    TCCR2B = BV(CS21) | BV(CS20);
    OCR2A  = 1*(F_CPU/32)/1000-1;
    TIMSK2 |= BV(OCIE2A);
    // USART 19200 Baud, 8 Data, 1 Stop, No Parity, Async. -------------------------------
    UBRR0  = F_CPU/16L/19200-1;
    UCSR0A = 0;
    UCSR0B = BV(RXCIE0) | BV(TXCIE0) | BV(RXEN0) | BV(TXEN0);
    UCSR0C = BV(UCSZ01) | BV(UCSZ00);
    // SPI Master, F_CPU÷16, MSB first, Sample on rising Edge, Mode 3 --------------------
    SPCR = BV(SPE)| BV(MSTR) | BV(CPOL) | BV(CPHA) | BV(SPR0);
    // Initialisierung LCD EA DOGM081 ----------------------------------------------------				
    SetPin(DSPL_RESET);   // Reset aufheben
    FreezeMs(50);         // 50ms warten
    LCDSendCmd(0x30);     // FUNCTION SET (DL=0, N=0, DH=0, IS21=00)
    LCDSendCmd(0x30);     // FUNCTION SET wdh.
    LCDSendCmd(0x31);     // FUNCTION SET (DL=0, N=0, DH=0, IS21=01)
    LCDSendCmd(0x14);     // BIAS SET (BS=0, FX=0)
    LCDSendCmd(0x74);     // CONTRAST SET (C[3210]=0100)
    LCDSendCmd(0x51);     // POWER/ICON/CONTRAST SET (Ion=0, Bon=0, C[54]=01)
    LCDSendCmd(0x6A);     // FOLLOWER CONTROL (Fon=1, Rab[210]=010)
    FreezeMs(250);        // Pause
    LCDSendCmd(0x0C);     // DISPLAY ON/OFF (D=1, C=0, B=0)
    LCDSendCmd(0x06);     // ENTRY MODE SET (I=1, S=0)
    LCDSendCmd(0x30);     // FUNCTION SET (DL=0, N=0, DH=0, IS21=00)
	// Benutzerdefinierte Zeichen für das LCD
	//
	//  '\x10'   '\x11'   '\x12'   '\x13'   '\x14'   '\x15'
	//
	//  .....    .....    #####    .....    .....    #####
	//  .....    .....    #####    #...#    #####    #...#
	//  .....    .....    #####    .#.#.    #....    #...#
	//  .....    .....    #####    ..#..    ###..    #...#
	//  .....    #####    #####    .#.#.    #....    #...#
	//  .....    #####    #####    #...#    #####    #...#
	//  .....    #####    #####    .....    .....    #...#
	//  #####    #####    #####    #####    #####    #####
	//
	static const UInt08 customChar[] PROGMEM =
	{   0x00, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x1F,
		0x00, 0x00, 0x00, 0x00, 0x1F, 0x1F, 0x1F, 0x1F,
		0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F,
		0x00, 0x11, 0x0A, 0x04, 0x0A, 0x11, 0x00, 0x1F,
		0x00, 0x1F, 0x10, 0x1C, 0x10, 0x1F, 0x00, 0x1F,
		0x1F, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x1F

	};	
    for (UInt08 i=0; i<CARDINALITY(customChar); i++)  // Benutzerdefinierte Zeichen setzen
    {   LCDSendCmd(0x40+i);
        LCDSendData(pgm_read_byte(&customChar[i]));
    }
    // Sonstiges -------------------------------------------------------------------------
    EnableInterrupts();
    doNext = DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_DoMainLoop(void)
{
    return doNext == DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Proceed(void)
{
    wdt_reset();
	ProceedLineVoltageMeasuring();
	ProceedPushButtons();
	ProceedIndicators();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Shutdown(void)
{
    while (BOARD_BusIsTransmitting());
    switch (doNext)
    {   case DO_ENTER_PGMLDR: EnterPgmLdr();        break;
        case DO_STARTUP:      Reset();              break;
        default:              ;
    }
    // Abschaltanweisungen ...
}

//////////////////////////////////////////////////////////////////////////////////////////
