﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Zeichenkette                                                                        //
//  ____________                                                                        //
//                                                                                      //
//  Copyright 2006 Reimar Grasbon                                                       //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Hinweis: Um die Datenstruktur Zeichenkette einsetzen zu können muss in Conf.h die   //
//           Konstante 'CFG_STRING_CAPACITY' definiert sein. Wertebereich 2 .. 250.     //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "String.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define DECIMAL_SEPARATOR '.'   // Punkt oder Komma

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen   /////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
bool
STRING_IsEmpty(const String* string)
{
    return string->length == 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STRING_Clear(String* string)
{
    string->buffer[0] = '\0';
    string->length    = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
UInt08
STRING_Length(const String* string)
{
    return string->length;
}

//////////////////////////////////////////////////////////////////////////////////////////
char
STRING_Last(const String* string)
{
    if (string->length == 0) return '\0';
    return string->buffer[string->length-1];
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STRING_Chop(String* string)
{
    if (string->length > 0) string->length--;
    string->buffer[string->length] = '\0';
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STRING_PrintChar(String* string, char c)
{
    if (string->length < CFG_STRING_CAPACITY)
    {   string->buffer[string->length] = c;
        string->length++;
        string->buffer[string->length] = '\0';
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  STRING_PrintText(): Fügt die angegebene Zeichenkette hinzu.                         //
//  ------------------                                                                  //
//                                                                                      //
//  Eingabe: · string: Adresse der String-Struktur.                                     //
//           · text:   Die auszugebende Zeichenkette (darf nicht NULL sein).            //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
STRING_PrintText(String* string, const char* text)
{
    while (*text != '\0')
    {   STRING_PrintChar(string, *text);
        text++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STRING_PrintTextP(String* string, const char* pgmText)
{
    while (true)
    {   char c = pgm_read_byte(pgmText);
        if (c == '\0') break;
        STRING_PrintChar(string, c);
        pgmText++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  STRING_PrintNumber(): Schreibt den übergebenen Wert als vorzeichenbehaftetete Ganz- //
//  --------------------  zahl in den String. Führende Nullen werden unterdrückt.       //
//                                                                                      //
//      Beispiele:  STRING_PrintNumber(1)        -> "1"                                 //
//                  STRING_PrintNumber(-1)       -> "-1"                                //
//                  STRING_PrintNumber(+32767)   -> "32767"                             //
//                  STRING_PrintNumber(-32768)   -> "-32768"                            //
//                  STRING_PrintNumber(0123)     -> "123"                               //
//                                                                                      //
//  Eingabe: · string: Adresse der String-Struktur.                                     //
//           · value:  Der auszugebende Wert: -32768 .. +32767.                         //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
STRING_PrintNumber(String* string, SInt16 value)
{
    if (value < 0)
    {   STRING_PrintChar(string, '-');
        if (value == -32768)
        {   STRING_PrintText(string, "32768");
            return;
        }
        value = -value;
    }
    // Vorzeichenlose 15bit Ganzzahl ausgeben
    bool zero  = false;
    UInt16 div = 10000;
    do
    {   UInt16 digit = (value/div)%10;
        if (zero || digit != 0)
        {   STRING_PrintChar(string, (char)('0'+(UInt08)digit));
            zero = true;
        }
        div /= 10;
        if (div == 1) zero = true;
    } while (div != 0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  STRING_PrintFixpoint(): Schreibt die angegebene, vorzeichenbehaftete Zahl als Fest- //
//  ----------------------  kommazahl in den String. Angegeben wird in 'precision' die  //
//  Anzahl der Nachkommastellen und damit die Interpretationsweise der Zahl in 'value'. //
//  'precision' kann Werte von 0 (Keine Teilung - also Eintel) bis 4 (Ausgabe als Zehn- //
//  tausendstel) annehmen.                                                              //
//                                                                                      //
//    Beispiel:                                                                         //
//               value |    0    |    1    |     2    |     3     |     4               //
//              -------+---------+---------+----------+-----------+-----------+         //
//                 123 |   "123"    "12.3"    "1.23"     "0.123"     "0.0123"           //
//                -123 |  "-123"   "-12.3"   "-1.23"    "-0.123"    "-0.0123"           //
//                  -1 |    "-1"    "-0.1"   "-0.01"    "-0.001"    "-0.0001"           //
//                   1 |     "1"     "0.1"    "0.01"     "0.001"     "0.0001"           //
//                   0 |     "0"     "0.0"    "0.00"     "0.000"     "0.0000"           //
//               12345 | "12345"  "1234.5"  "123.45"    "12.345"     "1.2345"           //
//                9999 |  "9999"   "999.9"   "99.99"     "9.999"     "0.9999"           //
//                                                                                      //
//  Eingabe: · value:     Auszugebender Wert als Festkommazahl.                         //
//           · precision: Festpunkt-Nachkommastellen: 0 .. 4.                           //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
STRING_PrintFixpoint(String* string, SInt16 value, UInt08 precision)
{
    if (value == -32768)
    {   STRING_PrintLit(string, "-32768");
    }
    else
    {   if (value < 0)
        {   STRING_PrintChar(string, '-');
            value = -value;
        }
        switch (precision)
        {   case 4:  if( value < 10000 ) STRING_PrintChar(string, '0');
            case 3:  if( value <  1000 ) STRING_PrintChar(string, '0');
            case 2:  if( value <   100 ) STRING_PrintChar(string, '0');
            case 1:  if( value <    10 ) STRING_PrintChar(string, '0');
            case 0:  ;
        }
        STRING_PrintNumber(string, value);
    }
    if (precision != 0)
    {   // Trennzeichen nach vorn propagieren
        STRING_PrintChar(string, DECIMAL_SEPARATOR);
        UInt08 length = STRING_Length(string);
        char* buffer  = STRING_Data(string);
        for (UInt08 i=length-1; i>length-precision-1; i--)
        {   char tmp    = buffer[i];
            buffer[i]   = buffer[i-1];
            buffer[i-1] = tmp;

        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  STRING_ToLatin1(): Transkodiert die UTF-8-Zeichenkette in 'string' zu Latin-1 (ISO  //
//  -----------------  8859-1). Da Latin-1 nie mehr Bytes als UTF-8 benötigt, findet    //
//  die Umwandung "in situ" statt. Unicode-Zeichen, die in Latin-1 nicht abgebildet     //
//  sind, werden durch den Wert 150 ersetzt.                                            //
//                                                                                      //
//  Eingabe: · string: String, dessen UTF-8 Zeichenkette in eine Latin-1 Zeichenkette   //
//                     umgewandelt wird.                                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
STRING_ToLatin1(String* string)
{
    UInt08 position  = 0;
    UInt08 byteCount = 0;
    UInt08 n         = 0;
    UInt08 length    = 0;
    UInt08 charCode  = 0;
    while (position < CFG_STRING_CAPACITY)
    {   UInt08 byte = (UInt08)string->buffer[position];
        if (byte == 0) break;
        position++;
        if (n == 0)
        {   // UTF-8 Startbyte
            if      ((byte & 0x80) == 0x00) byteCount = 1;
            else if ((byte & 0xE0) == 0xC0) byteCount = 2;
            else if ((byte & 0xF0) == 0xE0) byteCount = 3;
            else if ((byte & 0xF8) == 0xF0) byteCount = 4;
            else                            break;            // Kodierungsfehler
            if      (byteCount == 1) charCode = byte;         // Direkt ASCII
            else if (byteCount == 2) charCode = byte & 0x1F;  // 2-Byte Sequenz
        }
        else
        {   // UTF-8 Folgebyte: 10xxxxxx
            if ((byte & 0xC0) != 0x80) break; // Kodierungsfehler
            byte &= 0x3F;
            if (byteCount == 2 && charCode < 4) charCode = (UInt08)(charCode<<6)+byte;
            else                                charCode = 150;     // Ersatzzeichen
        }
        n++;
        if (n == byteCount)
        {   n = 0;
            string->buffer[length] = (char)charCode;
            length++;
        }
    }
    string->buffer[length] = 0;
    string->length = length;
}

//////////////////////////////////////////////////////////////////////////////////////////
const char*
STRING_ConstData(const String* string)
{
    return string->buffer;
}

//////////////////////////////////////////////////////////////////////////////////////////
char*
STRING_Data(String* string)
{
    return string->buffer;
}

//////////////////////////////////////////////////////////////////////////////////////////



